//       1         2         3         4         5         6         7         8         9        10        11        12        13
//3456789012345678901234567890123456789012345678901234567890123456789012345678901234567890123456789012345678901234567890123456789012
//**********************************************************************************************************************************

#include <algorithm>
#include <charconv>
#include <compare>
#include <cstdint>
#include <filesystem>
#include <fstream>
#include <iomanip>
#include <iostream>
#include <iterator>
#include <locale>
#include <map>
#include <optional>
#include <string>
#include <string_view>
#include <type_traits>
#include <vector>

#include <argparse/argparse.hpp>
#include <Networking.h>

namespace
{
    /// IPV6 128 bit decimal
    struct IPV6Decimal
    {
        uint64_t high;
        uint64_t low;
    };

    /// Ordering function for IPV6 128 bit decimal
    constexpr std::strong_ordering operator<=>(const IPV6Decimal& left, const IPV6Decimal& right)
    {
        const auto highCmp{ left.high <=> right.high };
        if (highCmp != std::strong_ordering::equal)
        {
            return highCmp;
        }
        return left.low <=> right.low;
    };

    // Define IPV4 and IPV6 to country maps
    using countryIPV4 = std::map<uint32_t, std::pair<uint32_t, std::string>>;
    using countryIPV6 = std::map<IPV6Decimal, std::pair<IPV6Decimal, std::string>>;

    // Define function declarations
    [[nodiscard]] std::pair<countryIPV4, countryIPV6> readCountryFile(const std::string& fileName);
    void parseAuthentication(const std::vector<std::string>& inputFiles, const uint16_t threshold, 
                             const std::pair<countryIPV4, countryIPV6>& countryIPs);
    [[nodiscard]] std::optional<uint32_t> convertIPV4ToDecimal(const std::string_view& ip);
    [[nodiscard]] std::optional<IPV6Decimal> convertIPV6ToDecimal(const std::string_view& ip);
    [[nodiscard]] std::string convertDecimalToIPV4(uint32_t decimal);
    [[nodiscard]] std::string convertDecimalToIPV6(IPV6Decimal decimal);
    template<const bool IPV4>
    [[nodiscard]] std::string resolveIP(const std::string& ip, const uint16_t port);
    [[nodiscard]] const std::string& lookupIPV4Country(const uint32_t ip, const countryIPV4& ipV4Countries);
    [[nodiscard]] const std::string& lookupIPV6Country(const IPV6Decimal& ip, const countryIPV6& ipV6Countries);
    [[nodiscard]] std::string_view trimWhitespace(std::string_view input);
}

int main(int argc, char ** argv)
{
    argparse::ArgumentParser options("SmarterMail Log Tool");

    // Declare the supported options.
    options.add_argument("-h", "--help").help("shows help message").default_value(false).implicit_value(true);
    options.add_argument("--country_file").help("the ip to country file");
    options.add_argument("--input_files").help("the input files").nargs(argparse::nargs_pattern::any);
    options.add_argument("--input_directory").help("the input directory");
    options.add_argument("--task").required().help("task to perform").default_value(std::string{ "authentication" });
    options.add_argument("--threshold").required().help("count threshold").default_value(2U).scan<'u', unsigned>();
    auto& inputGroup{ options.add_mutually_exclusive_group(true) };
    inputGroup.add_argument("--input_files");
    inputGroup.add_argument("--input_directory");

    try {
        options.parse_args(argc, argv);
    }
    catch (const std::exception& err) {
        std::cerr << err.what() << '\n';
        std::cerr << options;
        return EXIT_FAILURE;
    }

    if (options["--help"] == true)
    {
        std::cout << options;
        return EXIT_SUCCESS;
    }

    // Get the input files
    std::vector<std::string> inputFiles{};
    if (options.is_used("--input_files"))
    {
        // individual files
        inputFiles = options.get<std::vector<std::string>>("--input_files");
    }
    else
    {
        // an input directory
        const std::string inputDirectory{ options.get<std::string>("--input_directory") };
        for (auto const& dir_entry : std::filesystem::directory_iterator{ inputDirectory })
        {
            if (dir_entry.is_regular_file() && (dir_entry.path().extension() == ".log"))
            {
                inputFiles.emplace_back(dir_entry.path().string());
            }
        }
    }

    std::pair<countryIPV4, countryIPV6> countryIPs;
    if (options.is_used("--country_file"))
    {
        countryIPs = readCountryFile(options.get<std::string>("--country_file"));
    }

    // Perform the appropriate task
    const std::string task{ options.get<std::string>("--task") };
    if (task == std::string{ "authentication" })
    {
        const uint16_t threshold{ static_cast<uint16_t>(options.get<unsigned>("--threshold")) };
        parseAuthentication(inputFiles, threshold, countryIPs);
    }
    else
    {
        std::cerr << "SmarterMail Log Tool: Unknown Task: " << task << '\n';
        return EXIT_FAILURE;
    }
    return EXIT_SUCCESS;
}

namespace 
{
    /// Read in the country file which maps ip ranges to country
    /// @param fileName the name of the country file
    /// @return the ip V4 and ip V6 to country mapping
    [[nodiscard]] std::pair<countryIPV4, countryIPV6> readCountryFile(const std::string& fileName)
    {
        countryIPV4 ipV4Countries{};
        countryIPV6 ipV6Countries{};

        auto fileStream{ std::fstream{fileName, std::ios::in} };
        if (!fileStream)
        {
            std::cerr << "Could not open country file: " << fileName << '\n';
            return std::make_pair(std::move(ipV4Countries), std::move(ipV6Countries));
        }

        std::string line{};
        size_t lineNum{ 1U };

        // skip first header line
        std::getline(fileStream, line);

        // read line by line
        size_t ipV4Entries{ 0U };
        size_t ipV6Entries{ 0U };
        while (std::getline(fileStream, line))
        {
            ++lineNum;

            // get the beginning ip
            size_t pos{ line.find(',') };
            if (pos == std::string::npos)
            {
                continue;
            }
            const std::string_view begIP{ line.data(), pos };

            // get the ending ip
            size_t prevPos{ pos + 1U };
            pos = line.find(',', prevPos);
            if (pos == std::string::npos)
            {
                continue;
            }
            const std::string_view endIP{ line.data() + prevPos, pos - prevPos };

            // get the country
            prevPos = pos + 1U;
            pos = line.find(',', prevPos);
            if (pos == std::string::npos)
            {
                continue;
            }

            // get the country name
            prevPos = pos + 1U;
            pos = line.find(',', prevPos);
            if (pos == std::string::npos)
            {
                continue;
            }
            const std::string_view countryName{ line.data() + prevPos, pos - prevPos };

            const bool ipV4{ begIP.find('.') != std::string_view::npos };
            if (ipV4)
            {
                const auto begDecimal{ convertIPV4ToDecimal(begIP) };
                const auto endDecimal{ convertIPV4ToDecimal(endIP) };
                if (!begDecimal || !endDecimal)
                {
                    std::cerr << "Line: " << lineNum << " - Invalid Beg IPV4: " << begIP << " or End IPV4: " << endIP << '\n';
                    continue;
                }
                ipV4Countries.emplace(*begDecimal, std::make_pair(*endDecimal, countryName));
                ++ipV4Entries;
            }
            else
            {
                const auto begDecimal{ convertIPV6ToDecimal(begIP) };
                const auto endDecimal{ convertIPV6ToDecimal(endIP) };
                if (!begDecimal || !endDecimal)
                {
                    std::cerr << "Line: " << lineNum << " - Invalid Beg IPV6: " << begIP << " or End IPV6: " << endIP << '\n';
                    continue;
                }
                ipV6Countries.emplace(*begDecimal, std::make_pair(*endDecimal, countryName));
                ++ipV6Entries;
            }
        }

        std::cout << "Country File: " << fileName << " - IPV4 Entries: " << ipV4Entries << ", IPV6 Entries: "
                  << ipV6Entries << '\n';
        return std::make_pair(std::move(ipV4Countries), std::move(ipV6Countries));
    }


    /// Output the ips in the container
    /// @tparam IPContainerType the type of the ip container
    /// @param ipContainer the ip container
    /// @param threshold the threshold for the count
    /// @param countryIPs the country IPs
    template<typename IPContainerType>
    void outputIPs(const IPContainerType& ipContainer, const size_t threshold, 
                   const std::pair<countryIPV4, countryIPV6>& countryIPs)
    {
        for (const auto& [ip, count] : ipContainer)
        {
            if (count >= threshold)
            {
                constexpr bool ipV4{ std::is_same_v<typename IPContainerType::key_type, uint32_t> };
                std::string ipStr{};
                std::string hosts{};
                std::string country{};
                if constexpr (ipV4)
                {
                    ipStr = convertDecimalToIPV4(ip);
                    hosts = resolveIP<true>(ipStr, 25);
                    country = lookupIPV4Country(ip, countryIPs.first);
                }
                else
                {
                    ipStr = convertDecimalToIPV6(ip);
                    hosts = resolveIP<false>(ipStr, 25);
                    country = lookupIPV6Country(ip, countryIPs.second);
                }
                std::cout << std::setw(39) << std::left << ipStr << std::setw(6) << std::right << count << ' ' << std::setw(20)
                          << std::left << country << hosts << '\n';
            }
        }
    }

    /// Parse the SMTP logs for RSP 535 errors
    /// @param inputFiles the input log files to parse
    /// @param threshold the threshold count to report the ip
    /// @param countryIPs the ip to country mappings
    void parseAuthentication(const std::vector<std::string>& inputFiles, const uint16_t threshold, 
                             const std::pair<countryIPV4, countryIPV6>& countryIPs)
    {
        // read through files and get ips
        std::map<uint32_t, size_t> ipV4Authenticate;
        std::map<uint32_t, size_t> ipV4Blacklist;
        std::map<IPV6Decimal, size_t> ipV6Authenticate;
        std::map<IPV6Decimal, size_t> ipV6Blacklist;
        std::cout << "Parsing SMTP log files:\n";
        size_t fileNum{ 0U };
        for(const auto& inputFile : inputFiles)
        {
            if (inputFile.find("smtpLog") == std::string::npos)
            {
                // incorrect file
                continue;
            }

            ++fileNum;
            std::cout << "File: " << std::setw(3) << fileNum << " - " << inputFile << '\n';

            auto fileStream{ std::fstream{inputFile, std::ios::in} };
            if (!fileStream)
            {
                std::cerr << "Could not open SMTP log file: " << inputFile << '\n';
                continue;
            }

            // read line by line
            std::string line{};
            size_t lineNum{ 0U };
            while(std::getline(fileStream, line))
            {
                ++lineNum;
                const bool rsp535{ line.find("rsp: 535 Authentication failed") != std::string::npos };
                const bool blacklisted{ line.find("IP is blacklisted") != std::string::npos };
                if (rsp535 || blacklisted)
                {
                    // ip authentication failure so grab ip
                    const auto posOpenBracket{ line.find('[') };
                    const auto posCloseBracket{ line.find(']', posOpenBracket + 1U) };
                    if ((posOpenBracket == std::string::npos) || (posCloseBracket == std::string::npos))
                    {
                        std::cerr << "Ill formed line IP: " << lineNum << " line: " << line;
                        continue;
                    }

                    const std::string_view ip{ line.data() + posOpenBracket + 1U, posCloseBracket - posOpenBracket - 1U };
                    const bool ipV4{ ip.find('.') != std::string_view::npos };
                    if (ipV4)
                    {
                        const std::optional<uint32_t> ipDecimal{ convertIPV4ToDecimal(ip) };
                        if (!ipDecimal)
                        {
                            std::cerr << "Line: " << lineNum << " - Invalid IPV4: " << ip << '\n';
                            continue;
                        }
                        if (rsp535)
                        {
                            const auto [itr, success] { ipV4Authenticate.emplace(*ipDecimal, 0U) };
                            ++(itr->second);
                        }
                        else
                        {
                            const auto [itr, success] { ipV4Blacklist.emplace(*ipDecimal, 0U) };
                            ++(itr->second);
                        }
                    }
                    else
                    {
                        const std::optional<IPV6Decimal> ipDecimal{ convertIPV6ToDecimal(ip) };
                        if (!ipDecimal)
                        {
                            std::cerr << "Line: " << lineNum << " - Invalid IPV6: " << ip << '\n';
                            continue;
                        }
                        if (rsp535)
                        {
                            const auto [itr, success] { ipV6Authenticate.emplace(*ipDecimal, 0U) };
                            ++(itr->second);
                        }
                        else
                        {
                            const auto [itr, success] { ipV6Blacklist.emplace(*ipDecimal, 0U) };
                            ++(itr->second);
                        }
                    }
                }
            }
        }

        // Report the IP's
        std::cout << "\nAuthentication failure IPs:" << '\n';
        std::cout << std::setw(39) << std::left << "IP:" << std::setw(6) << std::left << "Count:" << std::setw(20)
                  << std::left << " Country:" << " Host Names:" << '\n';
        outputIPs(ipV4Authenticate, threshold, countryIPs);
        outputIPs(ipV6Authenticate, threshold, countryIPs);

        std::cout << "\nBlacklisted IPs:" << '\n';
        std::cout << std::setw(39) << std::left << "IP:" << std::setw(6) << std::left << "Count:" << std::setw(20)
                  << std::left << " Country:" << " Host Names:" << '\n';
        outputIPs(ipV4Blacklist, threshold, countryIPs);
        outputIPs(ipV6Blacklist, threshold, countryIPs);
    }
    
    /// Convert the IPV4 address to decimal
    /// @param ip the ip to convert
    /// @return the IPV4 address in decimal form if valid ip
    [[nodiscard]] std::optional<uint32_t> convertIPV4ToDecimal(const std::string_view& ip)
    {
        uint32_t decimalIP{ 0U };
        uint32_t multiplier{ static_cast<uint32_t>(256UL * 256UL * 256UL) };

        const std::string_view trimmedIP{ trimWhitespace(ip) };

        size_t posPrev{ 0U };
        auto posDot{ trimmedIP.find('.') };
        for (size_t octNum{ 0U }; octNum < 4U; ++octNum, multiplier /= 256UL)
        {
            // get octet
            const std::string_view octet{ trimmedIP.data() + posPrev, std::min(posDot, ip.size()) - posPrev};
            uint8_t value;
            auto [ptr, ec] {std::from_chars(octet.data(), octet.data() + octet.size(), value)};
            if ((ec == std::errc{}) && (ptr == octet.data() + octet.size()))
            {
                decimalIP += multiplier * static_cast<uint32_t>(value);
            }
            else
            {
                // invalid IPV4
                return {};
            }

            if (posDot == std::string_view::npos)
            {
                // finished with segments
                break;
            }

            // find next
            posPrev = posDot + 1U;
            posDot = trimmedIP.find('.', posPrev);
        }

        if (posDot != std::string_view::npos)
        {
            // invalid IPV4 since did not reach end of ip during parsing
            return {};
        }

        return decimalIP;
    }

    /// Convert the IPV6 address to decimal
    /// @param ip the ip to convert
    /// @return the IPV6 address in decimal form
    [[nodiscard]] std::optional<IPV6Decimal> convertIPV6ToDecimal(const std::string_view& ip)
    {
        IPV6Decimal decimalIP{ 0U, 0U };

        const std::string_view strippedIP{ trimWhitespace(ip) };

        // look for double ::
        std::string_view leftHextets{};
        std::string_view rightHextets{};
        const size_t posDS{ strippedIP.find("::") };
        if (posDS != std::string_view::npos)
        {
            // found :: so split into left and right
            leftHextets = std::string_view{ strippedIP.data(), posDS };
            rightHextets = std::string_view{ strippedIP.data() + posDS + 2U, strippedIP.size() - posDS - 2U};
        }
        else
        {
            // no ::, so all left hextets
            leftHextets = std::string_view{ strippedIP.data(), strippedIP.size() };
        }

        size_t numHextets{ 0U };

        // parse left hextets
        if (!leftHextets.empty())
        {
            uint64_t multiplier{ static_cast<uint64_t>(65536ULL * 65536ULL * 65536ULL) };
            size_t posPrev{ 0U };
            auto posSemi{ leftHextets.find(':') };
            bool firstPart{ true };
            // loop over all hextets with up to 8 possible
            for (size_t hexNum{ 0U }; hexNum < 8U; ++hexNum, multiplier /= 65536ULL)
            {
                if (hexNum == 4U)
                {
                    // reset for second part
                    multiplier = 65536ULL * 65536ULL * 65536ULL;
                    firstPart = false;
                }

                // get hextet
                const std::string_view hextet{ leftHextets.data() + posPrev,
                                              std::min(posSemi, leftHextets.size()) - posPrev};
                uint16_t value;
                auto [ptr, ec] {std::from_chars(hextet.data(), hextet.data() + hextet.size(), value,
                                                                16)};
                if ((ec == std::errc{}) && (ptr == hextet.data() + hextet.size()))
                {
                    ++numHextets;
                    if (firstPart)
                    {
                        decimalIP.high += multiplier * static_cast<uint64_t>(value);
                    }
                    else
                    {
                        decimalIP.low += multiplier * static_cast<uint64_t>(value);
                    }
                }
                else
                {
                    // invalid IPV6
                    return {};
                }

                if (posSemi == std::string::npos)
                {
                    // parsed all hextets
                    break;
                }

                // find next
                posPrev = posSemi + 1U;
                posSemi = leftHextets.find(':', posPrev);
            }

            if (posSemi != std::string_view::npos)
            {
                // invalid IPV6 since did not reach end of ip during parsing
                return {};
            }
        }

        // parse right hextets in reverse
        if (!rightHextets.empty())
        {
            uint64_t multiplier{ 1U };
            size_t posPrev{ rightHextets.size() - 1U };
            auto posSemi{ rightHextets.rfind(':', posPrev) };
            bool firstPart{ false };
            // loop over all hextets and early out if no more.  There are up to 7 possible
            for (size_t hexNum{ 0U }; hexNum < 7U; ++hexNum, multiplier *= 65536ULL)
            {
                if (hexNum == 4U)
                {
                    // reset for second part
                    multiplier = 1U;
                    firstPart = true;
                }

                // get hextet
                const std::string_view hextet{ rightHextets.data() + posSemi + 1U,
                                                   ((posSemi == std::string_view::npos) ? posPrev + 1U : posPrev - posSemi) };
                uint16_t value;
                auto [ptr, ec] {std::from_chars(hextet.data(), hextet.data() + hextet.size(), value,
                                                               16)};
                if ((ec == std::errc{}) && (ptr == hextet.data() + hextet.size()))
                {
                    ++numHextets;
                    if (firstPart)
                    {
                        decimalIP.high += multiplier * static_cast<uint64_t>(value);
                    }
                    else
                    {
                        decimalIP.low += multiplier * static_cast<uint64_t>(value);
                    }
                }
                else
                {
                    // invalid IPV6
                    return {};
                }

                if (posSemi == 0U)
                {
                    // invalid IPV6, since cannot begin with a : as this would indicate a ::: in original IP
                    return {};
                }

                if (posSemi == std::string::npos)
                {
                    // parsed all hextets
                    break;
                }

                // find previous
                posPrev = posSemi - 1U;
                posSemi = rightHextets.rfind(':', posPrev);
            }

            if (posSemi != std::string_view::npos)
            {
                // invalid IPV6 since did not reach end of ip during parsing
                return {};
            }
        }

        if ((posDS == std::string_view::npos) && (numHextets != 8U))
        {
            // invalid IPV6
            return {};
        }

        return decimalIP;
    }

    /// Convert the decimal form of the ip V4 address to string form
    /// @param decimal the decimal form of the IPV4 address
    /// @return the ip V4 address in string form
    [[nodiscard]] std::string convertDecimalToIPV4(uint32_t decimal)
    {
        std::string ip{};
        uint32_t multiplier{ static_cast<uint32_t>(256UL * 256UL * 256UL) };

        for (size_t octNum{ 0U }; octNum < 4U; ++octNum, multiplier /= 256UL)
        {
            const uint8_t value{ static_cast<uint8_t>(decimal/multiplier) };
            ip.append(std::to_string(value));

            // append separator
            if (octNum < 3U)
            {
                ip.push_back('.');
            }

            // remove the octet
            decimal -= static_cast<uint32_t>(value) * multiplier;
        }
        return ip;
    }

    /// Convert the decimal form of the ip V6 address to string form
    /// @param decimal the decimal form of the IPV6 address
    /// @return the ip V6 address in string form
    [[nodiscard]] std::string convertDecimalToIPV6(IPV6Decimal decimal)
    {
        std::string ip{};

        uint64_t multiplier{ static_cast<uint64_t>(65536ULL * 65536ULL * 65536ULL) };
        bool firstPart{ true };
        bool outputDS{ false };
        bool beginZeroSpan{ false };
        for (size_t hexNum{ 0U }; hexNum < 8U; ++hexNum, multiplier /= 65536UL)
        {
            if (hexNum == 4U)
            {
                // reset for second part
                multiplier = 65536ULL * 65536ULL * 65536ULL;
                firstPart = false;
            }

            const uint16_t value{ static_cast<uint16_t>(firstPart ? decimal.high / multiplier : decimal.low / multiplier) };
            if (value == 0U)
            {
                if (!outputDS)
                {
                    // no double semicolon
                    if (ip.empty())
                    {
                        // first entry so double semicolon
                        ip.append("::");
                    }
                    else
                    {
                        // only one semicolon since separator present
                        ip.push_back(':');
                    }
                    outputDS = true;
                    beginZeroSpan = true;
                }
                else if (!beginZeroSpan)
                {
                    // must output zeros since already did zero span
                    ip.push_back('0');
                }
            }
            else
            {
                // output a non zero hextet and end a zero span
                ip.append(std::to_string(value));
                beginZeroSpan = false;
            }

            // output separator
            if ((hexNum < 7U) && !beginZeroSpan)
            {
                ip.push_back(':');
            }

            // remove the value of this hextet
            if (firstPart)
            {
                decimal.high -= static_cast<uint64_t>(value) * multiplier;
            }
            else
            {
                decimal.low -= static_cast<uint64_t>(value) * multiplier;
            }
        }

        return ip;
    }

    /// Resolve the given IPV4 address into hostnames
    /// @tparam IPV4 whether the address is IPV4 or not
    /// @param ip the IP address 
    /// @param port the port number of the ip
    /// @return the hostnames separated by semicolons
    template<const bool IPV4>
    [[nodiscard]] std::string resolveIP(const std::string& ip, const uint16_t port)
    {
        std::string hostNames{};
        asio::io_service ioService{};
        asio::ip::tcp::resolver resolver{ ioService };
        asio::error_code ec{};
        std::conditional_t<IPV4, asio::ip::address_v4, asio::ip::address_v6> ipAddress;
        if constexpr (IPV4)
        {
            ipAddress = asio::ip::make_address_v4(ip, ec);
        }
        else
        {
            ipAddress = asio::ip::make_address_v6(ip, ec);
        };
        if (ec)
        {
            return hostNames;
        }

        // Resolve the IP host names
        const asio::ip::tcp::endpoint tcpEP{ ipAddress, port };
        auto itr{ resolver.resolve(tcpEP, ec) };
        const asio::ip::tcp::resolver::iterator itrEnd{};
        if (!ec)
        {
            size_t count{ 0U };
            for (; itr != itrEnd; ++itr, ++count)
            {
                if (count > 0U)
                {
                    hostNames.push_back(';');
                }
                hostNames.append(itr->host_name());
            }
        }
        // If host name is equal to input IP then no resolving
        if (hostNames == ip)
        {
            hostNames.clear();
        }
        return hostNames;
    }

    /// Lookup the IP V4 country based on the input country table
    /// @param ip the ip V4 address in decimal form
    /// @param ipV4Countries the ip V4 to countries mapping
    /// @return the country name
    [[nodiscard]] const std::string& lookupIPV4Country(const uint32_t ip, const countryIPV4& ipV4Countries)
    {
        static const std::string none{};
        if (ipV4Countries.empty())
        {
            return none;
        }
        const auto itr{ ipV4Countries.upper_bound(ip) };
        if (itr == std::cbegin(ipV4Countries))
        {
            return none;
        }

        const auto itrPrev{ std::prev(itr) };
        if ((itrPrev->first <= ip) && (itrPrev->second.first >= ip))
        {
            return itrPrev->second.second;
        }
        return none;
    }

    /// Lookup the IP V6 country based on the input country table
    /// @param ip the ip V6 address in decimal form
    /// @param ipV6Countries the ip V6 to countries mapping
    /// @return the country name
    [[nodiscard]] const std::string& lookupIPV6Country(const IPV6Decimal& ip, const countryIPV6& ipV6Countries)
    {
        static const std::string none{};
        if (ipV6Countries.empty())
        {
            return none;
        }
        const auto itr{ ipV6Countries.upper_bound(ip) };
        if (itr == std::cend(ipV6Countries))
        {
            return none;
        }

        const auto itrPrev{ std::prev(itr) };
        if ((itrPrev->first <= ip) && (itrPrev->second.first >= ip))
        {
            return itrPrev->second.second;
        }
        return none;
    }

    /// Trim the beginning and ending whitespace from the input
    /// @param input the string view to trim
    /// @return the stripped string view
    [[nodiscard]] std::string_view trimWhitespace(std::string_view input)
    {
        const auto isWhitespace{ [](const char c){
            static std::locale loc{};
            return std::isspace(c, loc);
        }};

        // trim the front
        while(!input.empty() && isWhitespace(input.front()))
        {
            input.remove_prefix(1U);
        }

        // trim the rear
        while (!input.empty() && isWhitespace(input.back()))
        {
            input.remove_suffix(1U);
        }
        return input;
    }
}

//**********************************************************************************************************************************
//3456789012345678901234567890123456789012345678901234567890123456789012345678901234567890123456789012345678901234567890123456789012
//       1         2         3         4         5         6         7         8         9        10        11        12        13
